@@ -1,27 +1,25 @@
 import copy
 from abc import abstractmethod
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
+from typing import TYPE_CHECKING, Dict, List, Optional, Union
 
 import openai
 from aiconfig.callback import CallbackEvent
 from aiconfig.default_parsers.parameterized_model_parser import ParameterizedModelParser
 from aiconfig.model_parser import InferenceOptions
 from aiconfig.util.config_utils import get_api_key_from_environment
-from aiconfig.util.params import (
-    resolve_parameters,
-    resolve_prompt,
-    resolve_prompt_string,
-    resolve_system_prompt,
-)
+from aiconfig.util.params import resolve_prompt, resolve_prompt_string, resolve_system_prompt
+from openai.types.chat import ChatCompletionMessage
 
-from aiconfig import schema
 from aiconfig.schema import (
     ExecuteResult,
-    ModelMetadata,
+    FunctionCallData,
     Output,
+    OutputDataWithToolCallsValue,
+    OutputDataWithValue,
     Prompt,
     PromptInput,
     PromptMetadata,
+    ToolCallData,
 )
 
 if TYPE_CHECKING:
@@ -60,7 +58,12 @@ class OpenAIInference(ParameterizedModelParser):
         event = CallbackEvent(
             "on_serialize_start",
             __name__,
-            {"prompt_name": prompt_name, "data": data, "parameters": parameters, "kwargs": kwargs},
+            {
+                "prompt_name": prompt_name,
+                "data": data,
+                "parameters": parameters,
+                "kwargs": kwargs,
+            },
         )
         await ai_config.callback_manager.run_callbacks(event)
         prompts = []
@@ -96,7 +99,7 @@ class OpenAIInference(ParameterizedModelParser):
             role = messsage["role"]
             if role == "user" or role == "function":
                 # Serialize User message as a prompt and save the assistant response as an output
-                assistant_response = None
+                assistant_response: Union[ChatCompletionMessage, None] = None
                 if i + 1 < len(conversation_data["messages"]):
                     next_message = conversation_data["messages"][i + 1]
                     if next_message["role"] == "assistant":
@@ -106,6 +109,20 @@ class OpenAIInference(ParameterizedModelParser):
 
                 input = messsage["content"] if role == "user" else PromptInput(**messsage)
 
+                assistant_output = []
+                if assistant_response is not None:
+                    output_data = build_output_data(assistant_response)
+                    metadata = {"raw_response": assistant_response}
+                    if assistant_response.get("role", None) is not None:
+                        metadata["role"] = assistant_response.get("role")
+                    assistant_output = [
+                        ExecuteResult(
+                            output_type="execute_result",
+                            execution_count=None,
+                            data=output_data,
+                            metadata=metadata,
+                        )
+                    ]
                 prompt = Prompt(
                     name=new_prompt_name,
                     input=input,
@@ -116,16 +133,7 @@ class OpenAIInference(ParameterizedModelParser):
                             "remember_chat_context": True,
                         }
                     ),
-                    outputs=[
-                        ExecuteResult(
-                            output_type="execute_result",
-                            execution_count=None,
-                            data=assistant_response,
-                            metadata={},
-                        )
-                    ]
-                    if assistant_response
-                    else [],
+                    outputs=assistant_output,
                 )
                 prompts.append(prompt)
             i += 1
@@ -137,9 +145,7 @@ class OpenAIInference(ParameterizedModelParser):
         await ai_config.callback_manager.run_callbacks(event)
         return prompts
 
-    async def deserialize(
-        self, prompt: Prompt, aiconfig: "AIConfigRuntime", params: Optional[Dict] = {}
-    ) -> Dict:
+    async def deserialize(self, prompt: Prompt, aiconfig: "AIConfigRuntime", params: Optional[Dict] = {}) -> Dict:
         """
         Defines how to parse a prompt in the .aiconfig for a particular model
         and constructs the completion params for that model.
@@ -150,9 +156,7 @@ class OpenAIInference(ParameterizedModelParser):
         Returns:
             dict: Model-specific completion parameters.
         """
-        await aiconfig.callback_manager.run_callbacks(
-            CallbackEvent("on_deserialize_start", __name__, {"prompt": prompt, "params": params})
-        )
+        await aiconfig.callback_manager.run_callbacks(CallbackEvent("on_deserialize_start", __name__, {"prompt": prompt, "params": params}))
         # Build Completion params
         model_settings = self.get_model_settings(prompt, aiconfig)
 
@@ -163,22 +167,17 @@ class OpenAIInference(ParameterizedModelParser):
             completion_params["messages"] = []
 
             # Add System Prompt
-            if "system_prompt" in model_settings:
+            if model_settings.get("system_prompt", None) is not None:
                 system_prompt = model_settings["system_prompt"]
                 if isinstance(system_prompt, dict):
                     # If system prompt is an object, then it should have content and role attributes
                     system_prompt = system_prompt["content"]
-                resolved_system_prompt = resolve_system_prompt(
-                    prompt, system_prompt, params, aiconfig
-                )
-                completion_params["messages"].append(
-                    {"content": resolved_system_prompt, "role": "system"}
-                )
+                resolved_system_prompt = resolve_system_prompt(prompt, system_prompt, params, aiconfig)
+                completion_params["messages"].append({"content": resolved_system_prompt, "role": "system"})
 
             # Default to always use chat context
             if not hasattr(prompt.metadata, "remember_chat_context") or (
-                hasattr(prompt.metadata, "remember_chat_context")
-                and prompt.metadata.remember_chat_context != False
+                hasattr(prompt.metadata, "remember_chat_context") and prompt.metadata.remember_chat_context != False
             ):
                 # handle chat history. check previous prompts for the same model. if same model, add prompt and its output to completion data if it has a completed output
                 for i, previous_prompt in enumerate(aiconfig.prompts):
@@ -189,25 +188,33 @@ class OpenAIInference(ParameterizedModelParser):
                     if aiconfig.get_model_name(previous_prompt) == aiconfig.get_model_name(prompt):
                         # Add prompt and its output to completion data. Constructing this prompt will take into account available parameters.
                         add_prompt_as_message(
-                            previous_prompt, aiconfig, completion_params["messages"], params
+                            previous_prompt,
+                            aiconfig,
+                            completion_params["messages"],
+                            params,
                         )
         else:
             # If messages are already specified in the model settings, then just resolve each message with the given parameters and append the latest message
             for i in range(len(completion_params.get("messages"))):
                 completion_params["messages"][i]["content"] = resolve_prompt_string(
-                    prompt, params, aiconfig, completion_params["messages"][i]["content"]
+                    prompt,
+                    params,
+                    aiconfig,
+                    completion_params["messages"][i]["content"],
                 )
 
         # Add in the latest prompt
         add_prompt_as_message(prompt, aiconfig, completion_params["messages"], params)
-        await aiconfig.callback_manager.run_callbacks(
-            CallbackEvent("on_deserialize_complete", __name__, {"output": completion_params})
-        )
+        await aiconfig.callback_manager.run_callbacks(CallbackEvent("on_deserialize_complete", __name__, {"output": completion_params}))
         return completion_params
 
     async def run_inference(
-        self, prompt: Prompt, aiconfig: "AIConfigRuntime", options: InferenceOptions, parameters
-    ) -> Output:
+        self,
+        prompt: Prompt,
+        aiconfig: "AIConfigRuntime",
+        options: InferenceOptions,
+        parameters,
+    ) -> List[Output]:
         """
         Invoked to run a prompt in the .aiconfig. This method should perform
         the actual model inference based on the provided prompt and inference settings.
@@ -228,7 +235,7 @@ class OpenAIInference(ParameterizedModelParser):
         )
 
         if not openai.api_key:
-            openai.api_key = get_api_key_from_environment("OPENAI_API_KEY")
+            openai.api_key = get_api_key_from_environment("OPENAI_API_KEY").unwrap()
 
         completion_data = await self.deserialize(prompt, aiconfig, parameters)
         # if stream enabled in runtime options and config, then stream. Otherwise don't stream.
@@ -244,21 +251,26 @@ class OpenAIInference(ParameterizedModelParser):
 
         response = openai.chat.completions.create(**completion_data)
         outputs = []
-
         if not stream:
             # # OpenAI>1.0.0 uses pydantic models for response
             response = response.model_dump(exclude_none=True)
-            response_without_choices = {
-                key: copy.deepcopy(value) for key, value in response.items() if key != "choices"
-            }
+
+            response_without_choices = {key: copy.deepcopy(value) for key, value in response.items() if key != "choices"}
             for i, choice in enumerate(response.get("choices")):
+                output_message = choice["message"]
+                output_data = build_output_data(output_message)
+
                 response_without_choices.update({"finish_reason": choice.get("finish_reason")})
+                metadata = {"raw_response": output_message, **response_without_choices}
+                if output_message.get("role", None) is not None:
+                    metadata["role"] = output_message.get("role")
+
                 output = ExecuteResult(
                     **{
                         "output_type": "execute_result",
-                        "data": choice["message"],
+                        "data": output_data,
                         "execution_count": i,
-                        "metadata": response_without_choices,
+                        "metadata": metadata,
                     }
                 )
 
@@ -269,12 +281,13 @@ class OpenAIInference(ParameterizedModelParser):
             for chunk in response:
                 # OpenAI>1.0.0 uses pydantic models. Chunk is of type ChatCompletionChunk; type is not directly importable from openai Library, will require some diffing
                 chunk = chunk.model_dump(exclude_none=True)
+                chunk_without_choices = {key: copy.deepcopy(value) for key, value in chunk.items() if key != "choices"}
                 # streaming only returns one chunk, one choice at a time (before 1.0.0). The order in which the choices are returned is not guaranteed.
                 messages = multi_choice_message_reducer(messages, chunk)
 
                 for i, choice in enumerate(chunk["choices"]):
                     index = choice.get("index")
-                    accumulated_message_for_choice = messages.get(index, {})
+                    accumulated_message_for_choice = messages.get(index, "")
                     delta = choice.get("delta")
 
                     if options and options.stream_callback:
@@ -283,20 +296,30 @@ class OpenAIInference(ParameterizedModelParser):
                     output = ExecuteResult(
                         **{
                             "output_type": "execute_result",
-                            "data": copy.deepcopy(accumulated_message_for_choice),
+                            "data": accumulated_message_for_choice,
                             "execution_count": index,
-                            "metadata": {"finish_reason": choice.get("finish_reason")},
+                            "metadata": chunk_without_choices,
                         }
                     )
                     outputs[index] = output
             outputs = [outputs[i] for i in sorted(list(outputs.keys()))]
 
+            # Now that we have the complete outputs, we can parse it into our object model properly
+            for output in outputs:
+                output_message = output.data
+                output_data = build_output_data(output.data)
+
+                metadata = {"raw_response": output_message}
+                if output_message.get("role", None) is not None:
+                    metadata["role"] = output_message.get("role")
+
+                output.data = output_data
+                output.metadata = {**output.metadata, **metadata}
+
         # rewrite or extend list of outputs?
         prompt.outputs = outputs
 
-        await aiconfig.callback_manager.run_callbacks(
-            CallbackEvent("on_run_complete", __name__, {"result": prompt.outputs})
-        )
+        await aiconfig.callback_manager.run_callbacks(CallbackEvent("on_run_complete", __name__, {"result": prompt.outputs}))
         return prompt.outputs
 
     def get_prompt_template(self, prompt: Prompt, aiconfig: "AIConfigRuntime") -> str:
@@ -312,7 +335,10 @@ class OpenAIInference(ParameterizedModelParser):
             return message.content or ""
 
     def get_output_text(
-        self, prompt: Prompt, aiconfig: "AIConfigRuntime", output: Optional[Output] = None
+        self,
+        prompt: Prompt,
+        aiconfig: "AIConfigRuntime",
+        output: Optional[Output] = None,
     ) -> str:
         if not output:
             output = aiconfig.get_latest_output(prompt)
@@ -321,15 +347,22 @@ class OpenAIInference(ParameterizedModelParser):
             return ""
 
         if output.output_type == "execute_result":
-            message = output.data
-            if message.get("content"):
-                return message.get("content")
-            elif message.get("function_call"):
-                return message.get("function_call")
-            else:
-                return ""
-        else:
-            return ""
+            output_data = output.data
+            if isinstance(output_data, str):
+                return output_data
+            if isinstance(output_data, OutputDataWithValue):
+                if isinstance(output_data.value, str):
+                    return output_data.value
+                # If we get here that means it must be of kind tool_calls
+                return output_data.model_dump_json(exclude_none=True, indent=2)
+            # Doing this to be backwards-compatible with old output format
+            # where we used to save the ChatCompletionMessage in output.data
+            if isinstance(output_data, ChatCompletionMessage):
+                if hasattr(output_data, "content") and output_data.content is not None:
+                    return output_data.content
+                elif output_data.function_call is not None:
+                    return str(output_data.function_call)
+        return ""
 
 
 class DefaultOpenAIParser(OpenAIInference):
@@ -341,24 +374,6 @@ class DefaultOpenAIParser(OpenAIInference):
         return self.model_id
 
 
-class GPT4Parser(DefaultOpenAIParser):
-    def __init__(self):
-        model_id = "gpt-4"
-        super().__init__(model_id)
-
-
-class GPT3TurboParser(DefaultOpenAIParser):
-    def __init__(self):
-        model_id = "gpt-3.5-turbo"
-        super().__init__(model_id)
-
-
-class ChatGPTParser(DefaultOpenAIParser):
-    def __init__(self):
-        model_id = "ChatGPT"
-        super().__init__(model_id)
-
-
 def reduce(acc, delta):
     acc = copy.deepcopy(acc)
 
@@ -376,9 +391,7 @@ def reduce(acc, delta):
     return acc
 
 
-def multi_choice_message_reducer(
-    messages: Union[Dict[int, dict], None], chunk: dict
-) -> Dict[int, dict]:
+def multi_choice_message_reducer(messages: Union[Dict[int, dict], None], chunk: dict) -> Dict[int, dict]:
     if messages is None:
         messages = {}
 
@@ -412,6 +425,8 @@ def refine_chat_completion_params(model_settings):
         "stop",
         "stream",
         "temperature",
+        "tools",
+        "tool_choice",
         "top_p",
         "user",
     }
@@ -424,9 +439,7 @@ def refine_chat_completion_params(model_settings):
     return completion_data
 
 
-def add_prompt_as_message(
-    prompt: Prompt, aiconfig: "AIConfigRuntime", messages: List, params=None
-):
+def add_prompt_as_message(prompt: Prompt, aiconfig: "AIConfigRuntime", messages: List, params=None):
     """
     Converts a given prompt to a message and adds it to the specified messages list.
 
@@ -460,9 +473,44 @@ def add_prompt_as_message(
     output = aiconfig.get_latest_output(prompt)
     if output:
         if output.output_type == "execute_result":
-            output_message = output.data
-            if output_message["role"] == "assistant":
+            assert isinstance(output, ExecuteResult)
+            output_data = output.data
+            role = output.metadata.get("role", None) or ("raw_response" in output.metadata and output.metadata["raw_response"].get("role", None))
+
+            if role == "assistant":
+                output_message = {}
+                content: Union[str, None] = None
+                function_call: Union[FunctionCallData, None] = None
+                if isinstance(output_data, str):
+                    content = output_data
+                elif isinstance(output_data, OutputDataWithValue):
+                    if isinstance(output_data.value, str):
+                        content = output_data.value
+                    elif output_data.kind == "tool_calls":
+                        assert isinstance(output_data, OutputDataWithToolCallsValue)
+                        function_call = output_data.value[len(output_data.value) - 1].function
+
+                output_message["content"] = content
+                output_message["role"] = role
+
+                # We should update this to use ChatCompletionAssistantMessageParam
+                # object with field `tool_calls. See comment for details:
+                # https://github.com/lastmile-ai/aiconfig/pull/610#discussion_r1437174736
+                if function_call is not None:
+                    output_message["function_call"] = function_call
+
+                name = output.metadata.get("name", None) or ("raw_response" in output.metadata and output.metadata["raw_response"].get("name", None))
+                if name is not None:
+                    output_message["name"] = name
+
                 messages.append(output_message)
+
+            # Doing this to be backwards-compatible with old output format
+            # where we used to save the ChatCompletionMessage in output.data
+            elif isinstance(output_data, ChatCompletionMessage):
+                if output_data.role == "assistant":
+                    messages.append(output_data)
+
     return messages
 
 
@@ -470,6 +518,59 @@ def is_prompt_template(prompt: Prompt):
     """
     Check if a prompt's input is a valid string.
     """
-    return isinstance(prompt.input, str) or (
-        hasattr(prompt.input, "data") and isinstance(prompt.input.data, str)
-    )
+    return isinstance(prompt.input, str) or (hasattr(prompt.input, "data") and isinstance(prompt.input.data, str))
+
+
+def build_output_data(
+    message: Union[ChatCompletionMessage, None],
+) -> Union[OutputDataWithValue, str, None]:
+    if message is None:
+        return None
+
+    output_data: Union[OutputDataWithValue, str, None] = None
+    if message.get("content") is not None and message.get("content") != "":
+        output_data = message.get("content")  # string
+    elif message.get("tool_calls") is not None:
+        tool_calls = []
+        for item in message.get("tool_calls"):
+            tool_call_type = item.get("type")
+            if tool_call_type != "function":
+                # It's possible that ChatCompletionMessageToolCall may
+                # support more than just function calls in the future
+                # so filter out other types of tool calls for now
+                continue
+
+            function = item.get("function")
+            tool_calls.append(
+                ToolCallData(
+                    id=item.get("id"),
+                    function=FunctionCallData(
+                        arguments=function.get("arguments"),
+                        name=function.get("name"),
+                    ),
+                    type=tool_call_type,
+                )
+            )
+        output_data = OutputDataWithToolCallsValue(
+            kind="tool_calls",
+            value=tool_calls,
+        )
+
+    # Deprecated, use tool_calls instead
+    elif message.get("function_call") is not None:
+        function_call = message.get("function_call")
+        tool_calls = [
+            ToolCallData(
+                id="function_call_data",  # value here does not matter
+                function=FunctionCallData(
+                    arguments=function_call["arguments"],
+                    name=function_call["name"],
+                ),
+                type="function",
+            )
+        ]
+        output_data = OutputDataWithToolCallsValue(
+            kind="tool_calls",
+            value=tool_calls,
+        )
+    return output_data
