@@ -12,9 +12,7 @@ from .util.file_path_utils import get_absolute_file_path_from_relative
 async def test_load_basic_chatgpt_query_config(set_temporary_env_vars):
     """Test loading a basic chatgpt query config"""
     config_relative_path = "aiconfigs/basic_chatgpt_query_config.json"
-    config_absolute_path = get_absolute_file_path_from_relative(
-        __file__, config_relative_path
-    )
+    config_absolute_path = get_absolute_file_path_from_relative(__file__, config_relative_path)
     config = AIConfigRuntime.load(config_absolute_path)
 
     data_for_inference = await config.resolve("prompt1")
@@ -24,9 +22,7 @@ async def test_load_basic_chatgpt_query_config(set_temporary_env_vars):
         "top_p": 1,
         "temperature": 1,
         "stream": False,
-        "messages": [
-            {"content": "Hi! Tell me 10 cool things to do in NYC.", "role": "user"}
-        ],
+        "messages": [{"content": "Hi! Tell me 10 cool things to do in NYC.", "role": "user"}],
     }
 
 
@@ -34,9 +30,7 @@ async def test_load_basic_chatgpt_query_config(set_temporary_env_vars):
 async def test_load_basic_dalle2_config(set_temporary_env_vars):
     """Test loading a basic Dall-E 2 config"""
     config_relative_path = "aiconfigs/basic_dalle2_config.json"
-    config_absolute_path = get_absolute_file_path_from_relative(
-        __file__, config_relative_path
-    )
+    config_absolute_path = get_absolute_file_path_from_relative(__file__, config_relative_path)
     config = AIConfigRuntime.load(config_absolute_path)
 
     data_for_inference = await config.resolve("panda_eating_dumplings")
@@ -53,9 +47,7 @@ async def test_load_basic_dalle2_config(set_temporary_env_vars):
 async def test_load_basic_dalle3_config(set_temporary_env_vars):
     """Test loading a basic Dall-E 3 config"""
     config_relative_path = "aiconfigs/basic_dalle3_config.json"
-    config_absolute_path = get_absolute_file_path_from_relative(
-        __file__, config_relative_path
-    )
+    config_absolute_path = get_absolute_file_path_from_relative(__file__, config_relative_path)
     config = AIConfigRuntime.load(config_absolute_path)
 
     data_for_inference = await config.resolve("panda_eating_dumplings")
@@ -72,9 +64,7 @@ async def test_load_basic_dalle3_config(set_temporary_env_vars):
 async def test_chained_gpt_config(set_temporary_env_vars):
     """Test loading a chained gpt config and resolving it, with chat context enabled"""
     config_relative_path = "aiconfigs/chained_gpt_config.json"
-    config_absolute_path = get_absolute_file_path_from_relative(
-        __file__, config_relative_path
-    )
+    config_absolute_path = get_absolute_file_path_from_relative(__file__, config_relative_path)
     config = AIConfigRuntime.load(config_absolute_path)
 
     data_for_inference1 = await config.resolve("prompt1")
@@ -119,9 +109,7 @@ async def test_resolve_system_prompt():
     Resolves a system prompt with a provided parameter
     """
     config_relative_path = "aiconfigs/system_prompt_parameters_config.json"
-    config_absolute_path = get_absolute_file_path_from_relative(
-        __file__, config_relative_path
-    )
+    config_absolute_path = get_absolute_file_path_from_relative(__file__, config_relative_path)
     config = AIConfigRuntime.load(config_absolute_path)
 
     data_for_inference = await config.resolve("prompt1", {"system": "skip odd numbers"})
